import java.util.Arrays;
/**
   A class that implements the ADT ring by using a resizable array.
   
   @author Frank M. Carrano 
   @version 4.0
 */
public class Ring<T> implements RingInterface<T>
{
	private T[] ring;			// Circular array of ring entries
	private int firstIndex;		// Marks the first entry added
	private int lastIndex;		// Marks the last entry added
	private int currentIndex;	// Marks the current entry
	private int numberOfEntries;
	private static final int DEFAULT_INITIAL_CAPACITY = 50;

	public Ring()
	{
		this(DEFAULT_INITIAL_CAPACITY);
	} // end default constructor

	public Ring(int initialCapacity)
	{
		numberOfEntries = 0;
		// The cast is safe because the new array contains null entries
		@SuppressWarnings("unchecked")
		T[] tempArray = (T[])new Object[initialCapacity]; // Unchecked cast
		ring = tempArray;
		firstIndex = 0;
		lastIndex = -1;
		currentIndex = 0;
	} // end constructor

	public Ring(T[] contents) 
	{
		ring = Arrays.copyOf(contents, contents.length);
		numberOfEntries = contents.length;
		firstIndex = 0;
		lastIndex = contents.length - 1;
		currentIndex = 0;
	} // end constructor

	public final void clear()
	{
		while (!isEmpty()) 
			remove();
		firstIndex = 0;
		lastIndex = -1;
		currentIndex = 0;
	} // end clear

	public void add(T newEntry)
	{
		ensureCapacity();
		lastIndex = (lastIndex + 1) % ring.length;
		ring[lastIndex] = newEntry; // Add at end of ring
		numberOfEntries++;
	} // end add

	// Doubles the size of the array ring if it is full
	private void ensureCapacity() 
	{
		// If array is full, double size of array
		if (numberOfEntries >= ring.length)
		{
			T[] oldRing = ring;
			int oldSize = ring.length;

			// The cast is safe because the new array contains null entries
			@SuppressWarnings("unchecked")
			T[] tempRing = (T[]) new Object[2 * oldSize];
			ring = tempRing;
			for (int index = 0; index < oldSize; index++)
			{
				ring[index] = oldRing[firstIndex];
				firstIndex = (firstIndex + 1) % oldSize;
			} // end for

			firstIndex = 0;
			lastIndex = oldSize - 1; 
		} // end if
	} // end ensureCapacity

	public T remove()
	{
		T result = null; // return value
		if (!isEmpty())
		{
			result = ring[firstIndex];
			ring[firstIndex] = null;
			firstIndex = (firstIndex + 1) % ring.length;
			numberOfEntries--;
		} // end if

		return result; // Return removed entry, or null if ring is empty
	} // end remove

	public T getCurrentEntry()
	{
		T result = null; // return value
		if (!isEmpty())
			result = ring[currentIndex];

		return result;  
	} // end getCurrentEntry

	public void advance()
	{        
		if (!isEmpty())
			currentIndex = (currentIndex + 1) % (lastIndex + 1);
	} // end advance

	public int getLength()
	{
		return numberOfEntries;
	} // end getLength

	public boolean isEmpty()
	{
		return numberOfEntries == 0;
	} // end isEmpty
}
