import java.util.Arrays;
/**
   A class that implements a bag of objects by using an array.
   The bag is never full.
   
   @author Frank M. Carrano
   @version 4.0
 */
public class ResizableArrayBag<T> implements BagInterface<T>
{
	private T[] bag; // Cannot be final due to doubling
	private int numberOfEntries;
	private static final int DEFAULT_CAPACITY = 25; // Initial capacity of bag

	/** Creates an empty bag whose initial capacity is 25. */
	public ResizableArrayBag() 
	{
		this(DEFAULT_CAPACITY);
	} // end default constructor

	/** Creates an empty bag having a given initial capacity.
		 @param capacity  The integer capacity desired. */
	public ResizableArrayBag(int capacity) 
	{
		// The cast is safe because the new array contains null entries
		@SuppressWarnings("unchecked")
		T[] tempBag = (T[])new Object[capacity]; // Unchecked cast
		bag = tempBag;
		numberOfEntries = 0;
	} // end constructor

	/** Creates a bag containing given entries.
		 @param contents  An array of objects. */
	public ResizableArrayBag(T[] contents) 
	{
		bag = Arrays.copyOf(contents, contents.length);
		numberOfEntries = contents.length;
	} // end constructor

	/** Adds a new entry to this bag.
		 @param newEntry  The object to be added as a new entry.
 		 @return  True if the addition is successful, or false if not. */
	public boolean add(T newEntry) 
	{
		ensureCapacity();
		bag[numberOfEntries] = newEntry;
		numberOfEntries++;

		return true;
	} // end add

	// Doubles the size of the array bag if it is full.
	private void ensureCapacity() 
	{
		if (numberOfEntries == bag.length)
			bag = Arrays.copyOf(bag, 2 * bag.length);
	} // end ensureCapacity

	/** Retrieves all entries that are in this bag.
		 @return  A newly allocated array of all the entries in the bag. */
	public T[] toArray() 
	{
		// The cast is safe because the new array contains null entries
		@SuppressWarnings("unchecked")
		T[] result = (T[])new Object[numberOfEntries]; // Unchecked cast

		for (int index = 0; index < numberOfEntries; index++) 
		{
			result[index] = bag[index];
		} // end for

		return result;
	} // end toArray

	/** Sees whether this bag is empty.
	    @return  True if this bag is empty, or false if not. */
	public boolean isEmpty() 
	{
		return numberOfEntries == 0;
	} // end isEmpty

	/** Gets the number of entries currently in this bag.
		 @return  The integer number of entries currently in the bag. */
	public int getCurrentSize() 
	{
		return numberOfEntries;
	} // end getCurrentSize

	/** Counts the number of times a given entry appears in this bag.
		 @param anEntry  The entry to be counted.
		 @return  The number of times anEntry appears in the bag. */
	public int getFrequencyOf(T anEntry) 
	{
		int counter = 0;

		for (int index = 0; index < numberOfEntries; index++) 
		{
			if (anEntry.equals(bag[index]))
			{
				counter++;
			} // end if
		} // end for

		return counter;
	} // end getFrequencyOf

	/** Tests whether this bag contains a given entry.
		 @param anEntry  The entry to locate.
		 @return  True if the bag contains anEntry, or false if not. */
	public boolean contains(T anEntry)
	{
		return getIndexOf(anEntry) > -1;
	} // end contains

	// Locates a given object within the array bag.
	// Returns the index of the object, if located,
	// or -1 otherwise.
	private int getIndexOf(T anEntry) 
	{
		int where = -1;
		boolean found = false;

		for (int index = 0; !found && (index < numberOfEntries); index++) 
		{
			if (anEntry.equals(bag[index]))
			{
				found = true;
				where = index;
			} // end if
		} // end for

		return where;
	} // end getIndexOf

	/** Removes all entries from this bag. */
	public void clear() 
	{
		while (!isEmpty()) 
			remove();
	} // end clear

	/** Removes one unspecified entry from this bag.
		 @return  Either the removed entry if the removal
				  was successful, or null if not. */
	public T remove()
	{
		T result = removeEntry(numberOfEntries - 1);

		return result;
	} // end remove

	/** Removes one occurrence of a given entry from this bag.
		 @param anEntry  The entry to be removed.
		 @return  True if the removal was successful, or null if not. */
	public boolean remove(T anEntry) 
	{
		int index = getIndexOf(anEntry);
		T result = removeEntry(index);

		return anEntry.equals(result);
	} // end remove

	// Removes and returns the array entry at a given index.
	// If no such entry exists, returns null.
	private T removeEntry(int givenIndex)
	{
		T result = null;

		if (!isEmpty() && (givenIndex >= 0))
		{
			result = bag[givenIndex]; // Entry to remove
			numberOfEntries--;
			bag[givenIndex] = bag[numberOfEntries]; // Replace entry to remove
													// with last entry
			bag[numberOfEntries] = null; // Remove reference to last entry
		} // end if

		return result;
	} // end removeEntry   

	public BagInterface<T> union(BagInterface<T> anotherBag)
	{
		BagInterface<T> unionBag = new ResizableArrayBag<>();
		ResizableArrayBag<T> otherBag = (ResizableArrayBag<T>)anotherBag;

		int index;

		// Add entries from this bag to the new bag
		for (index = 0; index < getCurrentSize(); index++)
			unionBag.add(bag[index]);

		// Add entries from the second bag to the new bag
		for (index = 0; index < otherBag.getCurrentSize(); index++)
			unionBag.add(otherBag.bag[index]);

		return unionBag;
	} // end union

	public BagInterface<T> intersection(BagInterface<T> anotherBag)
	{
		// The count of an item in the intersection
		// is the smaller of the count in each bag

		BagInterface<T> intersectionBag = new ResizableArrayBag<>();
		ResizableArrayBag<T> otherBag = (ResizableArrayBag<T>)anotherBag;
		BagInterface<T> copyOfAnotherBag = new ResizableArrayBag<>();

		int index;

		// Copy the second bag
		for (index = 0; index < otherBag.getCurrentSize(); index++) 
		{
			copyOfAnotherBag.add(otherBag.bag[index]);
		} // end for

		// Add to intersectionBag each item in this bag that matches an item
		// in anotherBag; once matched, remove it from the second bag      
		for (index = 0; index < getCurrentSize(); index++) 
		{
			if (copyOfAnotherBag.contains(bag[index]))
			{
				intersectionBag.add(bag[index]);
				copyOfAnotherBag.remove(bag[index]);
			} // end if
		} // end for

		return intersectionBag;
	} // end intersection

	public BagInterface<T> difference(BagInterface<T> anotherBag)
	{
		// The count of an item in the difference is the difference
		// of the counts in the two bags.

		BagInterface<T> differenceBag = new ResizableArrayBag<>();
		ResizableArrayBag<T> otherBag = (ResizableArrayBag<T>)anotherBag;

		int index;

		// Copy this bag  
		for (index = 0; index < getCurrentSize(); index++) 
		{
			differenceBag.add(bag[index]);
		} // end for

		// Remove the ones that are in anotherBag  
		for (index = 0; index < otherBag.getCurrentSize(); index++) 
		{
			if (differenceBag.contains(otherBag.bag[index]))
			{
				differenceBag.remove(otherBag.bag[index]);
			} // end if
		} // end for

		return differenceBag;
	} // end difference
	
} // end ResizableArrayBag
